# 零、前言

编写《精通 SciPy》的想法出现了，但是在发表《数值和科学计算学习》之后两个月。 在南卡罗来纳大学介绍这本书的过程中，我荣幸地向工程师，科学家和学生的不同受众介绍了本书的内容，他们每个人都有非常不同的研究问题和他们自己的一套首选计算资源 。 在演讲之后的几周中，我帮助了一些专业人士过渡到基于 SciPy 的环境。 在这些会议中，我们讨论了 SciPy 在幕后如何与他们已经使用的相同算法集（通常是相同代码）。 我们尝试了一些示例，并系统地获得了可比的性能。 我们立即看到基于健壮的脚本语言的通用环境的明显好处。 通过 SciPy 堆栈，我们发现了一种与同事，学生或雇主进行交流和共享结果的简便方法。 在所有情况下，切换到 SciPy 堆栈都为我们的小组提供了更快的设置，使新手可以快速掌握。

参与此过程的每个人都从*新手*变为*高级用户*，最终*很快掌握了* SciPy 堆栈。 在大多数情况下，与我一起工作的个人的科学背景使过渡变得无缝。 当他们能够将他们的研究背后的理论与所提供的解决方案进行对比时，掌握过程就变成了现实。 *啊哈时刻*总是在复制过程中得到仔细的指导和解释的过程中发生。

这恰恰是这本书背后的哲学。 我邀请您参加类似的会议。 每章都被设想为与具有一定科学需求的个人进行对话，这些需求以数值计算表示。 我们在一起发现了相关的例子，这些例子是解决这些问题的不同可能方法，其背后的理论以及每种方法的利弊。

写作过程遵循类似的路径，以获取引人入胜的示例集。 我与几个不同领域的同事进行了交谈。 每个部分都清楚地反映了这些交流。 在编写最具挑战性的章节（最后四个章节）时，这一点至关重要。 为了确保整本书具有相同的质量，并且始终尝试遵循一组严格的标准，需要花更长的时间才能使这些章节满意。 特别要提到的是 NASA Langley 研究中心的 Aaron Dutle，他帮助塑造了计算几何学这一章的一部分； Facebook 的数据分析师 Parsa Bakhtary，激发了本章中关于统计计算在数据分析中的应用的许多技术。 。

这是一次了不起的旅程，有助于加深我对数值方法的理解，拓宽了我在解决问题上的视野，并增强了我的科学成熟度。 我希望它对您有相同的影响。

# 本书涵盖的内容

第 1 章和“数值线性代数”概述了矩阵在解决科学计算中的作用。 对于理解后续章节的大多数过程和思想而言，这是至关重要的章节。 您将学习如何在 Python 中有效地构造和存储大型矩阵。 然后，我们着手研究对其的基本操作和操作，然后进行因式分解，矩阵方程的解以及特征值/特征向量的计算。

第 2 章，“插值和近似”开发了近似函数的先进技术，并将其应用于科学计算。 这是接下来两章的 segway。

第 3 章，“微分和积分”探索了产生函数导数的不同技术，更重要的是，如何通过积分过程有效地计算面积和体积。 这是两章专门介绍科学计算中数值方法核心的第一章。 第二部分也是第 5 章和“常微分方程的初值问题”的介绍，其中提到了常微分方程。

第 4 章，“非线性方程式和优化”是一门技术性很强的章节，在其中我们讨论根据所涉及函数的种类获得函数系统的根和极值的最佳方法。

第 5 章和“常微分方程的初值问题”是有关实际问题的五个章节中的第一章。 通过示例，我们向您展示解决微分方程组的最流行技术以及一些应用。

第 6 章和“计算几何”浏览了计算机科学这一领域中最重要的算法。

第 7 章和“描述性统计”是关于统计计算及其在数据分析中的应用的两章中的第一章。 在本章中，我们重点介绍概率和数据探索。

第 8 章，“推断和数据分析”是有关数据分析的第二章。 我们专注于统计推断，机器学习和数据挖掘。

第 9 章，“数字图像处理”是本书的最后一章。 在其中，我们探索了图像压缩，编辑，还原和分析的技术。

# 这本书需要什么

要处理这些示例并尝试使用本书的代码，您所需要的只是带有 SciPy 堆栈的 Python 的最新版本（2.7 或更高版本）：NumPy，SciPy 库，matplotlib，IPython，pandas 和 SymPy。 尽管整本书提供了独立安装所有这些程序的方法，但是我们建议您通过科学的 Python 发行版（例如 Anaconda）执行全局安装。

# 这本书是给谁的

尽管这本书和技术最终旨在供应用数学家，工程师和计算机科学家使用，但此处介绍的材料面向的是更广泛的读者。 所需要做的只是精通 Python，熟悉 iPython，对科学计算中的数值方法有一定的了解，以及对在科学，工程学或数据分析中开发严肃应用程序的浓厚兴趣。

# 约定

在本书中，您将找到许多可以区分不同类型信息的文本样式。 以下是这些样式的一些示例，并解释了其含义。

文本中的代码字，数据库表名称，文件夹名称，文件名，文件扩展名，路径名，伪 URL，用户输入和 Twitter 句柄如下所示：“我们可以通过使用 include 指令来包含其他上下文。”

任何命令行输入或输出的编写方式如下：

```py
In [7]: %time eigvals, v = spspla.eigsh(A, 5, which='SM')
CPU times: user 19.3 s, sys: 532 ms, total: 19.8 s
Wall time: 16.7 s
In [8]: print eigvals
[ 10.565523  10.663114  10.725135  10.752737  10.774503]

```

### 注意

警告或重要提示会出现在这样的框中。

### 提示

提示和技巧如下所示。

