# 零、前言

如今，科学家，工程师和定量数据分析师面临许多挑战。 数据科学家希望能够以最少的编程工作量就大型数据集进行数值分析。 他们还希望编写尽可能接近他们习惯的数学语言的可读，高效和快速的代码。 科学计算世界中有许多公认的解决方案。

C，C++ 和 Fortran 编程语言具有其优势，但它们不是交互式的，并且被许多人认为过于复杂。 常见的商业替代产品，例如 MATLAB，Maple 和 Mathematica，提供了功能强大的脚本语言，比任何通用编程语言都更加受限制。 还存在其他类似于 MATLAB 的开源工具，例如 R，GNU Octave 和 Scilab。 显然，它们也缺乏像 Python 这样的语言的功能。

Python 是一种流行的通用编程语言，已在科学界广泛使用。 您可以从 Python 轻松访问旧版 C，Fortran 或 R 代码。 它是面向对象的，并且被认为比 C 或 Fortran 更高。 它使您能够以最小的麻烦编写可读易懂的代码。 但是，它缺少现成的 MATLAB 等效项。 那就是 NumPy 的用武之地。这本书是关于 NumPy 和相关的 Python 库，例如 SciPy 和 matplotlib。

## 什么是 NumPy？

NumPy（数字 Python 的缩写）是一个用于科学计算的开源 Python 库。 它使您可以自然地使用数组和矩阵。 该库包含一长串有用的数学函数，包括一些用于线性代数，傅立叶变换和随机数生成例程的函数。 如果在系统上安装了 NumPy 线性代数模块，则使用线性代数库 LAPACK。 否则，NumPy 提供其自己的实现。 LAPACK 是一个著名的库，最初是用 Fortran 编写的，MATLAB 也依赖该库。 在某种程度上，NumPy 取代了 MATLAB 和 Mathematica 的某些功能，从而实现了快速的交互式原型制作。

我们不会从开发人员的角度讨论 NumPy，而是从用户的角度讨论 NumPy。 NumPy 是一个非常活跃的项目，并且有很多贡献者。 也许有一天，您将成为其中的一员！

## 历史

NumPy 基于其前身 Numeric。 Numeric 于 1995 年首次发布，现已弃用。 出于各种原因，Numeric 和 NumPy 均未将其纳入标准 Python 库。 但是，您可以单独安装 NumPy，这将在第 1 章，“NumPy 快速入门”中进行说明。

2001 年，受 Numeric 启发的许多人创建了 SciPy，这是一个开放源代码的科学计算 Python 库，其功能类似于 MATLAB，Maple 和 Mathematica。 大约在这个时候，人们对 Numeric 越来越不满意。 Numarray 被创建为 Numeric 的替代方案。 现在也已弃用。 在某些方面，它比“数字”更好，但工作方式却大不相同。 因此，SciPy 继续依赖于数值原理和数值数组对象。 按照最新和最好的新软件的惯例，Numarray 的出现导致围绕它的整个生态系统的开发，并提供了一系列有用的工具。

2005 年，SciPy 的早期撰稿人 Travis Oliphant 决定对这种情况采取一些措施。 他试图将 Numarray 的某些功能集成到 Numeric 中。 进行了彻底的重写，并最终在 2006 年发布了 NumPy 1.0。那时，NumPy 具有 Numeric 和 Numarray 的所有功能，以及更多。 有可用的工具来促进从 Numeric 和 Numarray 的升级。 推荐升级，因为不再积极支持 Numeric 和 Numarray。

最初，NumPy 代码是 SciPy 的一部分。 后来将其分离，SciPy 现在将其用于数组和矩阵处理。

## 本书涵盖的内容

第 1 章，“NumPy 快速入门”指导您完成在系统上安装 NumPy 并创建基本 NumPy 应用所需的步骤。

第 2 章，“从 NumPy 基础知识开始”，介绍了 NumPy 数组和基本原理。

第 3 章，“熟悉常用函数”教您最常用的 NumPy 函数-基本的数学和统计函数。

第 4 章，“为您带来便利的便利函数”告诉您一些使使用 NumPy 更加容易的函数。 这包括例如根据布尔条件选择数组的某些部分的函数。 您还将了解多项式以及如何操纵 NumPy 对象的形状。

第 5 章，“处理矩阵和函数”涵盖了矩阵和通用函数。 矩阵在数学中是众所周知的，并且在 NumPy 中也有它们的表示形式。 通用函数（ufuncs）可以逐元素或按标量作用于数组。 ufuncs 期望一组标量作为输入，并产生一组标量作为输出。

第 6 章，“进一步使用 NumPy 模块”，讨论了一些通用函数的基本模块。 这些函数通常可以映射到它们的数学对应项，例如加法，减法，除法和乘法。

第 7 章，“探究特殊例程”描述了一些更专门的 NumPy 函数。 作为 NumPy 的用户，我们有时会发现自己有特殊要求。 幸运的是，NumPy 可以满足我们的大多数需求。

第 8 章，“通过测试确保质量”教您如何编写 NumPy 单元测试。

第 9 章，“使用 matplotlib 绘图”深入介绍了 matplotlib，这是一个非常有用的 Python 绘图库。 NumPy 不能单独用于创建图形和绘图。 matplotlib 与 NumPy 很好地集成，并且具有与 MATLAB 相当的绘图功能。

第 10 章，“当 NumPy 不够用时 -- SciPy 和更多”的内容，将详细介绍 SciPy。 我们知道 SciPy 和 NumPy 在历史上是相关的。 正如*历史*部分所述，SciPy 是一个基于 NumPy 的高级 Python 科学计算框架。 可以与 NumPy 结合使用。

第 11 章，“玩转 Pygame”是本书的甜点。 您将学习如何使用 NumPy 和 Pygame 创建有趣的游戏。 在本章中，您还将对人工智能有所了解。

附录 A，“小测验答案”包含了本章中所有小测验问题的答案。

附录 B，“其他在线资源”包含指向 Python，数学和统计网站的链接。

附录 C，“NumPy 函数的参考”列出了一些有用的 NumPy 函数及其说明。

## 为什么使用 NumPy？

NumPy 代码比直接的 Python 代码干净得多，并且它试图完成相同的任务。 由于操作直接在数组和矩阵上进行，因此所需的循环更少。 许多便利和数学函数也使生活更轻松。 底层算法经受了时间的考验，并在设计时考虑了高性能。

与基本 Python 中的等效数据结构（例如列表列表）相比，NumPy 的数组存储效率更高。 数组 IO 也明显更快。 性能的提高与数组元素的数量成比例。 对于大型数组，使用 NumPy 确实值得。 可以将大小达 TB 的文件映射到数组，从而实现最佳的数据读写。

NumPy 数组的缺点是它们比普通列表更专业。 在数值计算的上下文之外，NumPy 数组用处不大。 NumPy 数组的技术细节将在后面的章节中讨论。

NumPy 的大部分都是用 C 编写的。这使 NumPy 的速度比纯 Python 代码快。 NumPy C API 也存在，它允许在 C 语言的帮助下进一步扩展功能。 C API 不在本书讨论范围之内。 最后，由于 NumPy 是开源的，因此您可以获得所有相关的优势。 价格是最低的，就像啤酒一样免费。 您不必每次有人加入您的团队时就担心许可证问题，也不需要升级软件。 源代码适用于所有人。 这当然有利于代码质量。

## NumPy 的局限性

如果您是 Java 程序员，则可能对 Jython（Python 的 Java 实现）感兴趣。 在这种情况下，我对你有个坏消息。 不幸的是，Jython 在 Java 虚拟机上运行并且无法访问 NumPy，因为 NumPy 的模块主要是用 C 编写的。您可以说 Jython 和 Python 是两个完全不同的领域，尽管它们确实实现了相同的规范。 《NumPy Cookbook 第二版》，*Packt Publishing*（由 *Ivan Idris* 撰写）中讨论了一些解决方法。

## 这本书需要什么

要尝试本书中的代码示例，您将需要最新的 NumPy 版本。 这意味着您还将需要 NumPy 支持的 Python 版本之一。 一些代码示例出于说明目的使用了 matplotlib。 并非严格要求 matplotlib 遵循这些示例，但是建议您也安装它。 上一章是关于 SciPy 的，其中有一个涉及 SciKits 的示例。

这是用于开发和测试代码示例的软件的列表：

*   Python 2.7
*   NumPy 1.9
*   SciPy 0.13
*   matplotlib 1.3.1
*   Pygame 1.9.1
*   IPython 2.4.1

不用说，您在计算机上不需要此软件和这些版本。 Python 和 NumPy 构成了您所需的绝对最小值。

## 这本书适合谁？

本书适用于寻求高质量，开放源代码数学库的科学家，工程师，程序员或分析师。 假定具备 Python 的知识。 此外，还需要对数学和统计学有一定的亲和力，或者至少是有兴趣的。 但是，我提供了简短的解释和指向学习资源的指针。

## 部分

在本书中，您会发现几个经常出现的标题（实战时间，刚刚发生了什么？，拥有围棋英雄和小测验）。

为了给出有关如何完成过程或任务的明确说明，我们使用以下部分。

## 实战时间 - 标题

1.  操作 1
2.  操作 2
3.  操作 3

说明通常需要一些额外的说明以确保它们有意义，因此在这些部分之后进行介绍。

### 刚刚发生了什么？

本节说明刚刚完成的任务或说明的工作方式。

您还将在本书中找到其他一些学习辅助工具。

### 小测验 - 标题

这些是简短的多项选择题，旨在帮助您测试自己的理解。

### 勇往直前 - 标题

这些都是实际的挑战，这些挑战使您有想法尝试所学到的东西。

## 约定

在本书中，您会发现许多可以区分不同类型信息的文本样式。 以下是这些样式的一些示例，并解释了其含义。

文本中的代码字如下所示：“注意`numpysum()`不需要`for`循环。”

代码块设置如下：

```py
def numpysum(n):
  a = numpy.arange(n) ** 2
  b = numpy.arange(n) ** 3
  c = a + b
  return c
```

当我们希望引起您对代码块特定部分的注意时，相关的行或项目将以粗体显示：

```py
reals = np.isreal(xpoints)
print "Real number?", reals
Real number? [ True  True  True  True False False False False]

```

任何命令行输入或输出的编写方式如下：

```py
>>>fromnumpy.testing import rundocs
>>>rundocs('docstringtest.py')

```

**新术语**和**重要词**用粗体显示。 您在屏幕上看到的单词，例如在菜单或对话框中，将以如下形式出现：“单击**下一步**按钮可将您移至下一个屏幕。”

#### 注意

警告或重要提示会出现在这样的框中。

#### 提示

提示和技巧如下所示。
