# 附录 C：NumPy 函数的参考

本附录包含有用的 NumPy 函数及其说明的列表。

*   `numpy.apply_along_axis`（`func1d, axis, arr, *args`）：沿`arr`的一维切片应用函数`func1d`。
*   `numpy.arange([start,] stop[, step,], dtype=None)`：创建一个 NumPy 数组，它在指定范围内均匀间隔。
*   `numpy.argsort(a, axis=-1, kind='quicksort', order=None)`：返回对输入数组进行排序的索引。
*   `numpy.argmax(a, axis=None)`：返回沿轴的最大值的索引。
*   `numpy.argmin(a, axis=None)`：返回沿轴的最小值的索引。
*   `numpy.argwhere(a)`：查找非零元素的索引。
*   `numpy.array(object, dtype=None, copy=True, order=None, subok=False, ndmin=0)`：从类似数组的序列（例如 Python 列表）创建 NumPy 数组。
*   `numpy.testing.assert_allclose((actual, desired, rtol=1e-07, atol=0, err_msg='', verbose=True)`：如果两个对象在指定的精度下不相等，则引发错误。
*   `numpy.testing.assert_almost_equal()`：如果两个数字在指定的精度下不相等，则引发异常。
*   `numpy.testing.assert_approx_equal()`：如果两个数字在某个有效数字下不相等，则引发异常。
*   `numpy.testing.assert_array_almost_equal()`：如果两个数组在指定的精度下不相等，则引发异常。
*   `numpy.testing.assert_array_almost_equal_nulp(x, y, nulp=1)`：将数组与其**最低精度单位**（**ULP**）。
*   `numpy.testing.assert_array_equal()`：如果两个数组不相等，则引发异常。
*   `numpy.testing.assert_array_less()`：如果两个数组的形状不同，并且第一个数组的元素严格小于第二个数组的元素，则会引发异常。
*   `numpy.testing.assert_array_max_ulp(a, b, maxulp=1, dtype=None)`：确定数组元素最多相差 ULP 的指定数量。
*   `numpy.testing.assert_equal()`：测试两个 NumPy 数组是否相等。
*   `numpy.testing.assert_raises()`：如果使用定义的参数调用的可调用对象未引发指定的异常，则失败。
*   `numpy.testing.assert_string_equal()`：断言两个字符串相等。
*   `numpy.testing.assert_warns()`：如果未引发指定的警告，则失败。
*   `numpy.bartlett(M)`：返回带有`M`点的 Bartlett 窗口。 此窗口类似于三角形窗口。
*   `numpy.random.binomial(n, p, size=None)`：从二项分布中抽取随机样本。
*   `numpy.bitwise_and(x1, x2[, out])`：计算数组的按位`AND`。
*   `numpy.bitwise_xor(x1, x2[, out])`：计算数组的按位`XOR`。
*   `numpy.blackman(M)`：返回一个具有`M`点的布莱克曼窗口，该窗口接近最佳值，并且比凯撒窗口差。
*   `numpy.column_stack(tup)`：堆叠以元组列形式提供的一维数组  。
*   `numpy.concatenate ((a1, a2, ...), axis=0)`：将数组序列连接在一起。
*   `numpy.convolve(a, v, mode='full')`：计算一维数组的线性卷积。
*   `numpy.dot(a, b, out=None)`：计算两个数组的点积。
*   `numpy.diff(a, n=1, axis=-1)`：计算给定轴的 N 阶差。
*   `numpy.dsplit(ary, indices_or_sections)`：沿着第三轴将数组拆分为子数组。
*   `numpy.dstack(tup)`：沿第三轴堆叠以元组形式给出的数组。
*   `numpy.eye(N, M=None, k=0, dtype=<type 'float'>)`：返回单位矩阵。
*   `numpy.extract(condition, arr)`：使用条件选择数组的元素。
*   `numpy.fft.fftshift(x, axes=None)`：将信号的零频率分量移到频谱的中心。
*   `numpy.hamming(M)`：返回带有`M`点的汉明窗口。
*   `numpy.hanning(M)`：返回具有`M`点的汉宁窗口。
*   `numpy.hstack(tup)`：水平堆叠以元组形式给出的数组。
*   `numpy.isreal(x)`：返回一个布尔数组，其中`True`对应于输入数组的实数（而不是复数）元素。
*   `numpy.kaiser(M, beta)`：对于给定的`beta`参数，返回带有`M`点的凯撒窗口。
*   `numpy.load(file, mmap_mode=None)`：从`.npy`，`.npz`，或腌制中加载 NumPy 数组或腌制对象。 内存映射的数组存储在文件系统中，不必完全加载到内存中。 这对于大型数组尤其有用。
*   `numpy.loadtxt(fname, dtype=<type 'float'>, comments='#', delimiter=None, converters=None, skiprows=0, usecols=None, unpack=False, ndmin=0)`：将文本文件中的数据加载到 NumPy 数组中。
*   `numpy.lexsort (keys, axis=-1)`：使用多个键进行排序。
*   `numpy.linspace(start, stop, num=50, endpoint=True, retstep=False, dtype=None)`：返回在间隔内均匀间隔的数字。
*   `numpy.max(a, axis=None, out=None, keepdims=False)`：沿轴返回数组的最大值。
*   `numpy.mean(a, axis=None, dtype=None, out=None, keepdims=False)`：沿给定轴计算算术平均值。
*   `numpy.median(a, axis=None, out=None, overwrite_input=False)`：沿给定轴计算中位数  。
*   `numpy.meshgrid(*xi, **kwargs)`：返回坐标向量的坐标矩阵。 例如：

    ```py
    In: numpy.meshgrid([1, 2], [3, 4])
    Out:
    [array([[1, 2],
            [1, 2]]), array([[3, 3],
            [4, 4]])]
    ```

*   `numpy.min(a, axis=None, out=None, keepdims=False)`：沿轴返回数组的最小值。
*   `numpy.msort(a)`：返回沿第一轴排序的数组的副本。
*   `numpy.nanargmax(a, axis=None)`：返回给定一个忽略 NaN 的轴的最大值的索引。
*   `numpy.nanargmin(a, axis=None)`：返回给定的轴的最小值索引，忽略 NaN。
*   `numpy.nonzero(a)`：返回非零数组元素的索引。
*   `numpy.ones(shape, dtype=None, order='C')`：创建指定形状和数据类型的 NumPy 数组，包含 1s。
*   `numpy.piecewise(x, condlist, funclist, *args, **kw)`：分段求值函数。
*   `numpy.polyder(p, m=1)`：将多项式微分为给定阶数。
*   `numpy.polyfit(x, y, deg, rcond=None, full=False, w=None, cov=False)`：执行最小二乘多项式拟合。
*   `numpy.polysub(a1, a2)`：减去多项式。
*   `numpy.polyval(p, x)`：以指定值求值多项式。
*   `numpy.prod(a, axis=None, dtype=None, out=None, keepdims=False)`：返回指定轴上数组元素的乘积  。
*   `numpy.ravel(a, order='C')`：展平数组，或在必要时返回副本。
*   `numpy.reshape(a, newshape, order='C')`：更改 NumPy 数组的形状  。
*   `numpy.row_stack(tup)`：逐行堆叠数组。
*   `numpy.save(file, arr)`：以 NumPy `.npy`格式将 NumPy 数组保存到文件中。
*   `numpy.savetxt(fname, X, fmt='%.18e', delimiter=' ', newline='\n', header='', footer='', comments='# ')`：将 NumPy 数组保存到文本文件。
*   `numpy.sinc(a)`：计算`sinc`函数。
*   `numpy.sort_complex(a)`：首先以实部，然后是虚部对数组元素进行排序。
*   `numpy.split(a, indices_or_sections, axis=0)`：将数组拆分为子数组。
*   `numpy.std(a, axis=None, dtype=None, out=None, ddof=0, keepdims=False)`：沿给定轴返回标准差。
*   `numpy.take(a, indices, axis=None, out=None, mode='raise')`：使用指定的索引从数组中选择元素。
*   `numpy.vsplit(a, indices_or_sections)`：将数组垂直拆分为子数组。
*   `numpy.vstack(tup)`：垂直堆叠数组。
*   `numpy.where(condition, [x, y])`：基于布尔条件从输入数组中选择数组元素。
*   `numpy.zeros(shape, dtype=float, order='C')`：创建指定形状和数据类型的 NumPy 数组，其中包含零。